//************************************************************************************************************************************//
#include "Smartcard.h"


//************************************************************************************************************************************//
enum sync_state_t { SYNC_START_STATE, SYNC_START_CONDITION, SYNC_START_FINISHED, SYNC_CLK_LOW, SYNC_CLK_HIGH, SYNC_PRE_STOP_CONDITION, SYNC_STOP_CONDITION, SYNC_STOP_FINISHED, SYNC_FINISHED };
enum state_t { START_STATE, FOUND_FIRST_FALLING_EDGE, FOUND_FIRST_RAISING_EDGE, SYNC_FOUND, READING_BITS, SENDING_BITS, PARITY_BIT, PARITY_ERROR, PARITY_READ, PRE_FINISHED, FINISHED };


//************************************************************************************************************************************//
SmartCardReader::SmartCardReader(uint8_t c7_io, uint8_t c2_rst, uint8_t c1_vcc, uint8_t card_present, uint8_t c3_clk, boolean card_present_invert) {
  _io_in_pin  = c7_io;
  _rstin_pin  = c2_rst;
  _cmdvcc_pin = c1_vcc;
  _off_pin    = card_present;
  _clk_pin    = c3_clk;

  _guardTime    = 2 * DEFAULT_ETU;
  _ignoreParity = true;
  _timeOutCB    = NULL;
  _off_invert   = card_present_invert;

}


//************************************************************************************************************************************//
void SmartCardReader::_init(frequency_t freq) 
{
  pinMode(_off_pin, INPUT);
  pinMode(_cmdvcc_pin, OUTPUT);
  pinMode(_rstin_pin, OUTPUT);

  // Start with reading from Card
  pinMode(_io_in_pin, INPUT);

  // Default value for CMDVCCn and RSTIN
  if(!cardInserted()) 
  {
    digitalWrite(_cmdvcc_pin, LOW);
  }
  digitalWrite(_rstin_pin, LOW);

  switch(freq) 
  {
    /*
    See Atmel32u4 - Specification at Page 90ff

    Frequency = Clock / (2 * N * (1 + OCR1A))
    N = Prescaler Factor: 1,8,64,256,1024

    8Mhz:
    Frequency = 8/(2*1(1+OCR1A)) = 8/(2*(1+OCR1A))
    1Mhz = 8/2*(1+3) //OCR1A = 3
    2Mhz = 8/2*(1+1) //OCR1A = 1
    4Mhz = 8/2*(1+0) //OCR1A = 0
    */
    case CLK_2MHZ:
      _initial_etu  = 186;
      _guardTime    = 2 * _initial_etu;
      _ocra1        = 1;
      break;
    case CLK_2DOT5MHZ:
      _initial_etu  = 148;
      _guardTime    = 2 * _initial_etu;
      _ocra1        = 2;
      break;
    case CLK_4MHZ:
      _initial_etu  = 93;
      _guardTime    = 2 * _initial_etu;
      _ocra1        = 0;
      break;
    case CLK_1MHZ:
      _initial_etu  = DEFAULT_ETU;
      _guardTime    = 2 * _initial_etu;
      _ocra1        = 3;
      break;
    default:
      // NO CLK generation ...
      _initial_etu  = 50;
      _ocra1        = 0;
      break;

  } // end of switch block      
      
  _etu = _initial_etu;
  _activated = false;
  _high_active = true;
  
  // Set Clock to Output
  pinMode(_clk_pin, OUTPUT);
  
  // Do we have to generate CLK, by our own?
  if (_ocra1 > 0) 
  {
    TCNT1=0;

    // Toggle OC1A on Compare Match
    TCCR1A = 0x00;
    bitSet(TCCR1A, COM1A0);

    // Clear Timer on Compare Match
    TCCR1B = 0x00;
    bitSet(TCCR1B, WGM12);

    //16Mhz
    // Set frequency (0 = 8Mhz, 1 = 4MHz, 2 = 2.5Mhz, 3 = 2MHz, 7 = 1MHz)
    //8Mhz
    // Set frequency (0 = 4Mhz, 1 = 2MHz, 2 = 1.33Mhz, 3 = 1MHz)
    OCR1A = _ocra1;

    // No prescaling
    bitSet(TCCR1B, CS10);
  } 
  else 
  {
    // We need to stop a previous active clk
    TCCR1A = 0x00;
    TCCR1B = 0x00;
    OCR1A = _ocra1;

    // Set default LOW CLK
    digitalWrite(_clk_pin, LOW);
  }
  _currentIFS = T1_IFS_DEFAULT;
}


//************************************************************************************************************************************//
void SmartCardReader::ignoreParityErrors(boolean in) 
{
  _ignoreParity = in;
}


//************************************************************************************************************************************//
uint8_t SmartCardReader::calcEDC(uint8_t startValue, uint8_t *buf, uint16_t size) 
{
  uint8_t xorvalue = startValue;
  for(size_t i=0; i<size; i++) 
  {
    xorvalue ^= buf[i];
  }
  return xorvalue;
}


//************************************************************************************************************************************//
void SmartCardReader::setTimeOutCallback(SmartCardVoidFuncPtr cb) 
{
  _timeOutCB = cb;
}


//************************************************************************************************************************************//
boolean SmartCardReader::cardInserted() 
{
	if (_off_invert)
  {
    if(digitalRead(_off_pin) == LOW)
    {
      digitalWrite(_cmdvcc_pin, HIGH);
      return true;
    }
    else
    {
      digitalWrite(_cmdvcc_pin, LOW);
      return false;
    }
  }
	return (digitalRead(_off_pin) == HIGH);
}


//************************************************************************************************************************************//
uint16_t SmartCardReader::activate(uint8_t* buf, uint16_t buf_size, frequency_t frequency) 
{
  uint16_t bytes_received = 0;
  uint8_t ts = 0;

  // Just for consistency reasons
  if (_activated) 
  {
    deactivate();
  }
  
  if(buf != NULL && buf_size > 0) 
  {
    _init(CLK_1MHZ);
    _activateHW();

    // Read TS
    ts = _receiveTSByte();

    // try to read more
    if(ts != 0) 
    {
      // include TS in ATR
      buf[bytes_received++] = ts;
      bytes_received += receiveBytes(&buf[bytes_received], buf_size-bytes_received);
    }
    _activated = (bytes_received > 0);
  }
  return (_activated)?bytes_received:0;
}


//************************************************************************************************************************************//
uint16_t SmartCardReader::receiveBytes(uint8_t* buf, uint16_t buf_size) 
{
  uint16_t bytes_received = 0;
  _timeout = false;

  if (buf != NULL && buf_size > 0) 
  {
    while (cardInserted() && !_timeout && bytes_received < buf_size) 
    {
      if (_receiveByte(buf+bytes_received, _max_wwt) != 0) 
      {
        // An error occured, while receiving ATR
        break;
      }
      bytes_received++;
    }
  }
  return bytes_received;
}


//************************************************************************************************************************************//
boolean SmartCardReader::sendBytes(uint8_t* buf, uint16_t count) 
{
  boolean parityError = false;
  if (buf != NULL && count > 0) 
  {
    for(uint16_t i=0; i < count && cardInserted(); i++) 
    {
      // _sendByte returns false in case parity error occured 
      if (!_sendByte(buf[i]) && !_ignoreParity) 
      {
        delayMicroseconds(_guardTime);
        _sendByte(buf[i]);
        parityError = true;
      }
      if (i < count -1) 
      {
        // Only wait between bytes and not at the end
        delayMicroseconds(_guardTime);
      } 
      else 
      {
        // Debug
      }
    }
  }
  return parityError;
}


//************************************************************************************************************************************//
void SmartCardReader::deactivate() 
{
  // Turn of CLK generation
  _init(CLK_NO_CLK);

  // Remove any previous found values
  _etu = _initial_etu;
  _activated = false;
  _high_active = true;
  _timeout = false;
}


/******************************************************************************
 * Private section
 ******************************************************************************/

//************************************************************************************************************************************//
 boolean SmartCardReader::_sendByte(uint8_t out) 
 {
  uint8_t state = START_STATE;
  unsigned long nextBit = 0;
  boolean currentBit = false;
  boolean parity = false;
  boolean parityErrorSignaled = false;
  
  // 8 Data bits + parity
  uint8_t bits_left = 9;
  
  while (cardInserted() && state != FINISHED) 
  {
    switch(state) 
    {
      case START_STATE:
        // Change _io_in_pin to OUTPUT
        pinMode(_io_in_pin, OUTPUT);
        digitalWrite(_io_in_pin, LOW);

        nextBit = micros() + _etu;
        state=SENDING_BITS;
        break;
      case SENDING_BITS:
        if (micros() >= nextBit) 
        {
          currentBit = bitRead(out, 9 - bits_left);
          if (currentBit) 
          {
            parity = !parity;
          }  
        
          digitalWrite(_io_in_pin, ((currentBit)?HIGH:LOW));
          nextBit += _etu;
          bits_left--;
          
          if (1 >= bits_left) 
          {
            state = PARITY_BIT;
          }
        }
        break;
      case PARITY_BIT:
        if (micros() >= nextBit) 
        {
          // Write Parity Bit
          digitalWrite(_io_in_pin, ((parity)?HIGH:LOW));
          
          // Create Parity Bit          
          nextBit += _etu;
          state = PARITY_ERROR;
        }
        break;
      case PARITY_ERROR:
        if (micros() >= nextBit) 
        {
          // Wait for parity signal from card
          digitalWrite(_io_in_pin, HIGH);
          pinMode(_io_in_pin, INPUT);                 

          // Read 
          nextBit+= _etu/2;          
          state=PARITY_READ;
        }
        break;
      case PARITY_READ:
        if (micros() >= nextBit) 
        {
          parityErrorSignaled = (digitalRead(_io_in_pin) != HIGH);

          // Stop Sending Bity
          state=FINISHED;
        }
        break;
      default:
        state = FINISHED;
        break;
    }
  }
  return !parityErrorSignaled;
}   


//************************************************************************************************************************************//
int SmartCardReader::_receiveByte(uint8_t* buf, unsigned long timeout) 
{
  boolean startbit_found = false;
  
  // Calculate maximum wait time 
  unsigned long endTime = micros() + timeout;
  int           result = 0;
  unsigned long nextBitTime = 0;
  
  // Wait for start bit
  while (cardInserted() && !startbit_found) 
  {

    // First check if timeout occured
    if (micros() >= endTime) 
    {
      // Timeout
      _timeOutoccured();
      result = -1;
      break;
    }

    // We are waiting for the falling edge of the start bit
    if (digitalRead(_io_in_pin) != HIGH) 
    {
      // found it
      startbit_found = true;

      // Set time for first bit
      nextBitTime = micros() + _etu + _etu / 2;
    }
  }
  
  if (startbit_found) 
  {
    _receiveDataBits(buf, nextBitTime, 8);
  }
  return result;
}


//************************************************************************************************************************************//
void SmartCardReader::_receiveDataBits(uint8_t* buf, unsigned long startTime, uint8_t count) 
{
  // We also need to read the parity bit
  uint8_t bits_left = count + 1;
  unsigned long nextBitTime = startTime;
  boolean parity = true;
  uint8_t state = READING_BITS;
  
  // Init buffer
  *buf = 0;

  while (cardInserted() && state != FINISHED && bits_left > 0) 
  {
    switch(state) 
    {
      case READING_BITS:
        if(micros() >= nextBitTime) 
        {
          *buf = *buf >> 1;

          if (digitalRead(_io_in_pin) != LOW) 
          {
            *buf |= 0x80;
            parity = !parity;
          } 
          else 
          {
            // Be sure to delete highest bit
            *buf &= 0x7F;
          }
          bits_left--;


          // Next is the Parity Bit
          if (bits_left <= 1) 
          {
            state = PARITY_BIT;
          }          
          // Set arrival time for next data bit
          nextBitTime += _etu;
        }
        break;
      case PARITY_BIT:
        // Wait until next bit arrives
        if(micros() >= nextBitTime) 
        {
          // Signal Parity Error, if requested
          if (!_ignoreParity && ((parity && digitalRead(_io_in_pin) != HIGH) || (!parity && digitalRead(_io_in_pin) != LOW)) ) 
          {
            state = PARITY_ERROR;
            nextBitTime += _etu / 2;
          } 
          else 
          {
            nextBitTime += _etu + _etu / 2;
            state = PRE_FINISHED;
          }
        }          
        break;
      case PARITY_ERROR:
        // Wait until parity error window arrives
        if(micros() >= nextBitTime) 
        {
          if (!_ignoreParity) 
          {
            // Signal Parity Error
            pinMode(_io_in_pin, OUTPUT);
            digitalWrite(_io_in_pin, LOW);
            delayMicroseconds(_etu);
            digitalWrite(_io_in_pin, LOW);
            pinMode(_io_in_pin, INPUT);
          } 
          else 
          {          
            delayMicroseconds(_etu);
          }
          nextBitTime += _etu / 2;
          state = PRE_FINISHED;
        }
        break;
      case PRE_FINISHED:
        // Just wait time to fly by
        if (micros() >= nextBitTime) 
        {
          state = FINISHED;
        }
        break;
      default:
        // Just be sure to bail out of while loop
        state = FINISHED;
        break;
    }
  }
}


//************************************************************************************************************************************//
uint8_t SmartCardReader::_receiveTSByte() 
{
  unsigned long start = 0;

  // Wait max 40.000 * etu
  unsigned long endTime = micros() + MAX_WAIT_TIME;

  // Init receiving state machine
  int state = START_STATE;
  
  // As we know TS, we preset bits_left and result, with predefined values
  int      bits_left = 6;
  uint8_t  result    = 0;
    
  // timeout did not occur and CARD still inserted
  while (cardInserted() && state != FINISHED && bits_left) 
  {
    // Check for timeout
    if(micros() >= endTime) 
    {
      result = 0;
      _timeOutoccured();
      break;
    }  
     switch (state) 
     {
       case START_STATE:
         if (digitalRead(_io_in_pin) != HIGH) 
         {
           start = micros();
           state = FOUND_FIRST_FALLING_EDGE;
         }
         break;
       case FOUND_FIRST_FALLING_EDGE:
         if (digitalRead(_io_in_pin) != LOW) 
         {
           state = FOUND_FIRST_RAISING_EDGE;
         }
         break;
       case FOUND_FIRST_RAISING_EDGE:
         if (digitalRead(_io_in_pin) != HIGH) 
         {
           // Calculate _etu
           _etu = ((micros() - start) / 3);

           // Calculate initial _guardtime (Wi * etu) (Wi = 2)
           _guardTime = 2 * _etu;

           // 960 * D * Wi ( D = 1, Wi = 2)
           _wwt       = 1920 * _etu; 

           // WWT + D * 480 etus ( D = 1 )
           // Just be a little more patient
           _max_wwt   = _wwt + 480 * DEFAULT_ETU;
           
           state = SYNC_FOUND;
         }
         break;
       case SYNC_FOUND:
         // Wait half etu, to read values in the middle of the etu
         _receiveDataBits(&result, micros() + _etu/2, bits_left);

         // Add predefined ATR BITS (First 2 are always high) 
         // Used for ETU calculation
         result |= 0x3;
         state = FINISHED;
         break;
       default:
    	   // We never should end up here ...
    	   result = 0;
         state = FINISHED;     
         break;
     }
  }
  return result;
}


//************************************************************************************************************************************//
//
// T1 functions
//
void SmartCardReader::sendAPDU(uint8_t nodeAdress, uint8_t* control, size_t size) 
{
  // Check all parameter
  static uint8_t iBlockCounter = 0;

  if (size == 0 || (size > 0 && control != NULL)) 
  {
    // Allocate maximum size iblock (chaining)
    uint8_t buffer[T1_OVERHEAD_SIZE];

    // Node Adress is always the same
    buffer[T1_OVERHEAD_NAD] = nodeAdress;

    // Do any needed sending
    int pkgCount = size / _currentIFS + 1;

    for(int i = 0; i < pkgCount; i++) 
    {
      // Do any needed chaining
      buffer[T1_OVERHEAD_PCB] = (i < pkgCount - 1) ? 0x20 : 0x00;

      // Sequence Number update needed
      if (iBlockCounter%2) 
      {
        buffer[T1_OVERHEAD_PCB] |= 0x40;
      }

      // Size field (Last package ??)
      buffer[T1_OVERHEAD_LEN] = (i < pkgCount - 1) ? _currentIFS : size % _currentIFS;

      // Send it out into the wild ...
      _sendT1Block(buffer, &control[i * _currentIFS]);

      iBlockCounter++;

      #warning AKNOWLEDGEMENT FROM CARD MISSING ???
    }
  }
}


//************************************************************************************************************************************//
void SmartCardReader::_sendT1Block(uint8_t* header, uint8_t* userData) 
{
  // First calc EDC
  // Calc EDC for block (Header (Without EDC itself) + opt. User Data)
  header[T1_OVERHEAD_EDC] = calcEDC(0, header, T1_OVERHEAD_SIZE - 1);
  if(header[T1_OVERHEAD_LEN] > 0) 
  {
    header[T1_OVERHEAD_EDC] = calcEDC(header[T1_OVERHEAD_EDC], userData, header[T1_OVERHEAD_LEN]);
  }

  // Send Header
  sendBytes(header, T1_OVERHEAD_SIZE - 1);

  if(header[T1_OVERHEAD_LEN] > 0) 
  {
    // Send User Data
    sendBytes(userData, header[T1_OVERHEAD_LEN]);
  }

  // Send EDC
  sendBytes(&header[T1_OVERHEAD_EDC], 1);
}



//************************************************************************************************************************************//
//
// HELPER FUNCTIONS
//
//************************************************************************************************************************************//
void SmartCardReader::dumpHEX(uint8_t* values, uint16_t size) 
{
  if (values != NULL && size > 0)
  {
    char ascii[17];
    for(uint16_t row=0; row<(size + 15)/16; row++) 
    {
      // Print Address
      if (row==0)
      {
        Serial.print("0");
      }
      Serial.print(row * 16, HEX);
      Serial.print("|");

      // Prefill ascii
      for(int i=0; i<16; i++)
      {
        ascii[i] = '.';
      }
      ascii[16] = (char)0x00;
      // colums
      for(uint16_t pos=row*16; pos<(row + 1) * 16; pos++ ) 
      {
        if(pos < size) 
        {
          if(values[pos] < 0x10) 
          {
            Serial.print("0");
          }
          Serial.print(values[pos], HEX);
          if(isPrintable(values[pos])) 
          {
            ascii[pos - row*16] = (char)values[pos];
          }
        } 
        else 
        {
          Serial.print("  ");
        }
        Serial.print(" ");
      }
      Serial.print("'");
      Serial.print(ascii);
      Serial.println("'");
    }
  }
}


//************************************************************************************************************************************//
void SmartCardReader::_activateHW() 
{
  // Start activate sequence (RSTIN high)
  digitalWrite( _rstin_pin, HIGH );

  // Wait some time
  delayMicroseconds(100);

  // Start activate sequence (RSTIN high)
  digitalWrite(_cmdvcc_pin, HIGH);

  // Wait at least t3
  delayMicroseconds(100);

  // Revert RST Signal
  digitalWrite( _rstin_pin, LOW );

  // Get over t5
  delayMicroseconds(150);

  // This should trigger a ATR ...
  digitalWrite( _rstin_pin, HIGH );
}


//************************************************************************************************************************************//
void SmartCardReader::_timeOutoccured()
{
  _timeout = true;
  if (_timeOutCB != NULL) 
  {
    _timeOutCB();
  }
}  
