#ifndef __SMARTCARD_H__
#define __SMARTCARD_H__


#if defined(ARDUINO) && ARDUINO >= 100
  #include <Arduino.h>
#else
  #include "WProgram.h"
#endif


//************************************************************************************************************************************//
typedef enum {
  CLK_1MHZ,
  CLK_2MHZ,
  CLK_2DOT5MHZ,
  CLK_4MHZ,
  CLK_NO_CLK,
} frequency_t;


// Callback func pointer
typedef void (*SmartCardVoidFuncPtr)(void);

// Default Config
#define DEFAULT_ETU                372L

// Maximum time of etu's to wait
#define MAX_WAIT_TIME            40000

// Maximum BYTES in ATR
#define MAX_ATR_BYTES 32


//************************************************************************************************************************************//
// T1 Protocol
// Maximum size for T1 data field (INF)
#define T1_INF_MAX_BYTES 0xFE
#define T1_IFS_DEFAULT   0x20

// Size of NAD, PCB, LEN and EDC
#define T1_OVERHEAD_NAD  0
#define T1_OVERHEAD_PCB  1
#define T1_OVERHEAD_LEN  2
#define T1_OVERHEAD_EDC  3
#define T1_OVERHEAD_SIZE T1_OVERHEAD_EDC + 1


//************************************************************************************************************************************//
class SmartCardReader
{
  public:
    //************************************************************************************************************************************//
    /**
     * Creates a basic smart card reader object, with mandatory
     * pins assigned to establish smart card communication.
     * Constructor to define communication pins
     * c7_io               - IO line, connected to Card (C7)
     * c2_rst              - RST for Card (C2)
     * c1_vcc              - Vcc for Card (C1)
     * card_present        - Card present - HIGH - Card available - LOW - Card removed (Can be changed by card_present_invert)
     * c3_clk              - clk signal to SC (C3) - (timer1 / pin9 used)
     * card_present_invert - Use inverted off signal (LOW Card available - HIGH Card removed)
     */
    SmartCardReader(uint8_t c7_io, uint8_t c2_rst, uint8_t c1_vcc, uint8_t card_present, uint8_t c3_clk, boolean card_present_invert=false);
    
    
    //************************************************************************************************************************************//
    /**
     * Is a smartcard inserted?
     * return true if card is inserted, otherwise false.
     */
    boolean cardInserted();
    
    
    //************************************************************************************************************************************//
    /**
     * Start activation sequence and return ATR result in buf (buf should be at least 32 Bytes to be able to
     * receive maximum ATR response.
     *
     * buf        : pointer to the data buffer, which should take the ATR
     * buf_size   : size of buffer 
     * frequency  : used to activate card (Only applies when asyncron cards are used
     *              (Defaults to 1 MHz as in the previous implementation)
     *
     * return number of received bytes, or 0 if error occured (timeout etc.) 
     */
    uint16_t activate(uint8_t* buf, uint16_t buf_size, frequency_t frequency=CLK_1MHZ);
    
    
    //************************************************************************************************************************************//
    /**
     * Start deactivation sequence
     */
    void deactivate();
    
    
    //************************************************************************************************************************************//
    /**
     * Don't care about parity errors, as some cards screw it up on purpose as they use odd parity ;-)
     */
    void ignoreParityErrors(boolean in);
    
    
    //************************************************************************************************************************************//
    /**
     * Set callback, to be noticed if time out occurs ...
     */
    void setTimeOutCallback(SmartCardVoidFuncPtr cb);
    
    
    //************************************************************************************************************************************//
    /**
     * Used to calculate ADPU crc value
     *
     * startValue should normaly be 0, only if sequend calls on a data structure are made the curent value 
     *            is used as input for the calculation of the subsequend call.
     * buf        pointer to the APDU command
     * size       number of bytes used for EDC calculation
     *
     * returns the current ECU value.
     */
    uint8_t calcEDC(uint8_t startValue, uint8_t *buf, uint16_t size);

    
    //************************************************************************************************************************************//
    void sendAPDU(uint8_t nodeAdress, uint8_t* control, size_t size);


    //************************************************************************************************************************************//
    /**
     * Send data to smartcard
     *
     * Return true, if the receiver signaled a parity error
     */
    boolean sendBytes(uint8_t* buf, uint16_t count);

    
    //************************************************************************************************************************************//
    /**
     * Receive data from SmartCard until buffer full or timeout
     *
     * Return number of received bytes
     */    
    uint16_t receiveBytes(uint8_t* buf, uint16_t buf_size);

    
    //************************************************************************************************************************************//
    // Just some debug function 
    void dumpHEX(uint8_t* values, uint16_t size);


  private:  
    //************************************************************************************************************************************//
    // Internal receive functions
    int _receiveByte(uint8_t* buf, unsigned long timeout);
    
    
    //************************************************************************************************************************************//
    uint8_t _receiveTSByte();
    
    
    //************************************************************************************************************************************//
    void _receiveDataBits(uint8_t* buf, unsigned long startTime, uint8_t count=8);
    

    //************************************************************************************************************************************//
    boolean _sendByte(uint8_t out);
    
    
    //************************************************************************************************************************************//
    // Send Block via T1 protocol
    void _sendT1Block(uint8_t* header, uint8_t* userData);

    
    //************************************************************************************************************************************//
    // Curent Value used as IFS for communication
    uint8_t _currentIFS;
    

    //************************************************************************************************************************************//
    // Activate the hardware
    void _activateHW();

    
    //************************************************************************************************************************************//
    /**
     * Init PINs and start CLK, if defined
     */
    void _init(frequency_t freq);

    
    //************************************************************************************************************************************//
    void _timeOutoccured();

    
    //************************************************************************************************************************************//
    // PIN configuration
    uint8_t _io_in_pin;  // (INPUT/OUTPUT) : I/O pin to communication with Smart Card
    uint8_t _rstin_pin;  // (OUTPUT)       : LOW: reset active.  HIGH: inactive.
    uint8_t _cmdvcc_pin; // (OUTPUT)       : LOW: Activation sequence initiated HIGH: No activation in progress
    uint8_t _off_pin;    // (INPUT)        : HIGH: Card present LOW: No card present
    uint8_t _clk_pin;    // (OUTPUT)       : CLK signal for Smart Card (NOT_A_PIN, when no CLK generation)


    //************************************************************************************************************************************//
    // Internal configuration
    boolean _ignoreParity;   // We don't care about screwd up parity from the smartcard
	boolean _off_invert;
    uint16_t _etu;
    uint16_t _initial_etu;
    uint8_t  _ocra1;
    SmartCardVoidFuncPtr _timeOutCB;
    
    // Some timing constants
    unsigned long _guardTime;
    
    // Work Waiting Time (960 * D * Wi) etus
    unsigned long _wwt;
    
    // Maximum Waiting time (WWT + (D * 480)) etus
    unsigned long _max_wwt;
    
    // Internal state vars
    boolean  _activated;
    boolean  _high_active;
    boolean  _timeout;
};

#endif // __SMARTCARD_H__
